
# coding: utf-8

# In[ ]:


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import matplotlib.pyplot as plt
import torchvision.models as models
import torchvision.transforms as transforms
import scipy.io as sio
import math
import PIL.Image as Image
import numpy.random as random
import numpy as np
from scipy.ndimage.interpolation import zoom
from torch.autograd import Variable
import sys
import pydensecrf.densecrf as dcrf
from pydensecrf.utils import unary_from_labels, create_pairwise_bilateral, create_pairwise_gaussian
from numbers import Number
from pydensecrf import eigen
#cimport eigen


vgg16 = models.vgg16(pretrained=True)


# In[ ]:



get_ipython().magic('matplotlib inline')
def imshow(img):
    npimg = img.numpy()
    npimg = np.transpose(npimg, (1,2,0))
    plt.imshow(npimg)

def imshowVariable(img):
    img_data = img.data.squeeze()
    mean = torch.Tensor([0.485, 0.456, 0.406]).view(3,1,1)
    std= torch.Tensor([0.229, 0.224, 0.225]).view(3,1,1)
    img_data = img_data.mul(std.expand_as(img_data))
    img_data = img_data.add(mean.expand_as(img_data))
    imshow(img_data)


# In[ ]:


class ScaleNonPreserving(object):
    """Rescales the input PIL.Image to the given 'size'.
    'size' will be the size of the smaller edge.
    For example, if height > width, then image will be
    rescaled to (size * height / width, size)
    size: size of the smaller edge
    interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        return img.resize((self.size, self.size), self.interpolation)



#accepts numpy array as input
def zoomImage(img, size):
    w, h, nc = img.shape
    if (w <= h and w == size) or (h <= w and h == size):
        return img
    if w < h:
        return zoom(img, (size/w, size/w, 1))
    else:
        return zoom(img, (size/h, size/h, 1))

def scaleImage(img, size):
    w, h = img.size
    if (w <= h and w == size) or (h <= w and h == size):
        return img
    if w < h:
        ow = size
        oh = int(size * h / w)
        return img.resize((ow, oh), Image.ANTIALIAS)
    else:
        oh = size
        ow = int(size * w / h)
        return img.resize((ow, oh), Image.ANTIALIAS)

def scaleImageNonPreserving(img, size):
    w, h = img.size
    ow = size
    oh = size
    return img.resize((ow, oh), Image.ANTIALIAS)


    
def scaleImageBox(img,box, size):
    w, h = img.size
    if (w <= h and w == size) or (h <= w and h == size):
        return img
    if w < h:
        ow = size
        oh = int(size * h / w)
        box[:,1:5].mul_(size/w)
        return img.resize((ow, oh), Image.ANTIALIAS), box.ceil()
    else:
        oh = size
        ow = int(size * w / h)
        box[:,1:5].mul_(size/h)
        return img.resize((ow, oh), Image.ANTIALIAS), box.ceil()
    


"""Converts a torch.*Tensor of range [0, 1] and shape C x H x W
or numpy ndarray of dtype=uint8, range[0, 255] and shape H x W x C
to a PIL.Image of range [0, 255]
"""

def toPilImage(pic):
    npimg = pic
    mode = None
    if not isinstance(npimg, np.ndarray):
        npimg = pic.mul(255).byte().numpy()
        npimg = np.transpose(npimg, (1, 2, 0))

    if npimg.shape[2] == 1:
        npimg = npimg[:, :, 0]
        mode = "L"

    return Image.fromarray(npimg, mode=mode)


"""Converts a PIL.Image or numpy.ndarray (H x W x C) in the range
[0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
"""

def toTensor(pic):
    if isinstance(pic, np.ndarray):
        # handle numpy array
        img = torch.from_numpy(pic).permute(2,0,1)
    else:
        # handle PIL Image
        img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
        # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
        if pic.mode == 'YCbCr':
            nchannel = 3
        else:
            nchannel = len(pic.mode)
        img = img.view(pic.size[1], pic.size[0], nchannel)
        # put it from HWC to CHW format
        # yikes, this transpose takes 80% of the loading time/CPU
        img = img.transpose(0, 1).transpose(0, 2).contiguous()
    return img.float().div(255)



"""Crops the given PIL.Image at a random location to have a region of
the given size. size can be a tuple (target_height, target_width)
or an integer, in which case the target will be of a square shape (size, size)
"""

def randomCrop(img, size):
    w,h = img.size
    if w == size and h == size:
        return img

    x1 = random.randint(0, w - size)
    y1 = random.randint(0, h - size)
    return img.crop((x1, y1, x1 + size, y1 + size))


def randomCropBox(img, box, size):
    w,h = img.size
    if w == size and h == size:
        return img

    x1 = random.randint(0, w - size)
    y1 = random.randint(0, h - size)
    
    box[:,1:3].sub_(x1)
    box[:,3:5].sub_(y1)
    box = box.clamp(0,223).int()
    return img.crop((x1, y1, x1 + size, y1 + size)), box

def displayBoxedImage(img, bbox, num):
    cur_label = bbox.int()
    if ~isinstance(img, np.ndarray):
        cur_data = np.asarray(img)
    else:
        cur_data = img
    mask = np.zeros(cur_data.shape, dtype='uint8')
    starty = cur_label[num][1]
    endy = cur_label[num][2]
    startx = cur_label[num][3]
    endx = cur_label[num][4]

    mask[startx:endx, starty:endy,:] = 1
    boxedImg = mask*cur_data
    plt.imshow(boxedImg)

    
        
"""Given mean: (R, G, B) and std: (R, G, B),
will normalize each channel of the torch.*Tensor, i.e.
channel = (channel - mean) / std
"""
def normalize(tensor, mean, std):
    # TODO: make efficient
    for t, m, s in zip(tensor, mean, std):
        t.sub_(m).div_(s)
    return tensor


def displayMaskedImage(img, mask):
    if ~isinstance(img, np.ndarray):
        cur_data = np.asarray(img)
    else:
        cur_data = img

    if isinstance(mask, Variable):
        mask = mask.data.cpu().numpy()
    elif ~isinstance(mask, np.ndarray):
        mask = mask.cpu().numpy()
    
    #mask = (mask-mask.min())/(mask.max()-mask.min())
    mask = np.tile(mask.reshape(img.shape[0],img.shape[1],1), (1,1,3) )
    maskedImage = mask*cur_data
    plt.imshow(maskedImage/maskedImage.max())


# In[ ]:


numTopics = 21
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Sequential(*list(vgg16.features.children())[0:4])
        #output size 224x224
        self.pool1 = nn.Sequential(*list(vgg16.features.children())[4:5])
        self.conv3 = nn.Sequential(*list(vgg16.features.children())[5:9])
        #output size 112x112
        self.pool2 = nn.Sequential(*list(vgg16.features.children())[9:10])
        self.conv5 = nn.Sequential(*list(vgg16.features.children())[10:16])
        #output size = 56x56
        self.pool3 = nn.Sequential(*list(vgg16.features.children())[16:17])
        self.conv8 = nn.Sequential(*list(vgg16.features.children())[17:23])
        #output size 28x28
        self.pool4 = nn.Sequential(*list(vgg16.features.children())[23:24])
        self.conv11 = nn.Sequential(*list(vgg16.features.children())[24:30])
        #output size 14x14
        
        #Apply to output of conv11
        self.conv12 = nn.ConvTranspose2d(512, 512, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
        self.bn12 = nn.BatchNorm2d(512)
        self.nlt12 = nn.LeakyReLU(.2)
        #output size 28x28
        #concatenate conv8's ouput here
        #Apply to concat of conv8 and nlt12
        self.conv13 = nn.Conv2d(1024, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.bn13 = nn.BatchNorm2d(512)
        self.nlt13 = nn.LeakyReLU(.2)
        #output size 28x28
        #Apply to nlt13
        self.conv14 = nn.ConvTranspose2d(512, 256, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
        self.bn14 = nn.BatchNorm2d(256)
        self.nlt14 = nn.LeakyReLU(.2)
        #output size 56x56
        
        #concatenate conv5 output here
        #Apply to concat of conv5 and nlt14
        self.conv15 = nn.Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.bn15 = nn.BatchNorm2d(512)
        self.nlt15 = nn.LeakyReLU(.2)
        #output size 56x56
        self.conv16 = nn.Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.bn16 = nn.BatchNorm2d(512)
        self.nlt16 = nn.LeakyReLU(.2)
        #output size 56x56
        """#Apply to nlt15
        self.conv16 = nn.ConvTranspose2d(256, 128, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
        self.bn16 = nn.BatchNorm2d(128)
        self.nlt16 = nn.LeakyReLU(.2)
        #output size 112x112
        
        #concatenate conv3 output here
        #Apply to concat of conv3 and nlt16
        self.conv17 = nn.Conv2d(256, 128, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        self.bn17 = nn.BatchNorm2d(128)
        self.nlt17 = nn.LeakyReLU(.2)
        #output size 112x112"""
        
        self.conv18 = nn.Conv2d(512, numTopics, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
        #output size 56x56
        
        

    def forward(self, x):
        out1 = self.conv1(x)
        #output size 64x224x224
        out1_5 = self.pool1(out1)
        #output size 64x112x112
        out2 = self.conv3(out1_5)
        #output size 128x112x112
        out3 = self.conv5(self.pool2(out2))
        #output size 256x56x56
        out4 = self.conv8(self.pool3(out3))
        #output size 512x28x28
        out5 = self.conv11(self.pool4(out4))
        #output size 512x14x14
        out6 = self.nlt12(self.bn12(self.conv12(out5)))
        #output size 512x28x28
        out7 = torch.cat([out4, out6], 1)
        #output size 1024x28x28
        out8 = self.nlt13(self.bn13(self.conv13(out7)))
        #output size 512x28x28
        out9 = self.nlt14(self.bn14(self.conv14(out8)))
        #output size 256x56x56
        out10 = torch.cat([out3, out9], 1)
        #output size 512x56x56
        out11 = self.nlt15(self.bn15(self.conv15(out10)))
        #output size 256x56x56
        out12 = self.nlt16(self.bn16(self.conv16(out11)))
        #output size 128x112x112
        #out13 = torch.cat([out2, out12],1)
        #output size 256x112x112
        #out14 = self.nlt17(self.bn17(self.conv17(out13)))
        #output size 128x112x112
        #out15 = torch.cat([out1_5, out14],1)
        #output size 192x112x112"""
        x = self.conv18(out12)
        x = torch.clamp(x,-3,3)
        #output size numTopicsx112x112
        x = torch.exp(x)
        temp = x.sum(1).expand_as(x)
        x = x/temp
        return x
net = Net()
ct = 0
for param in net.parameters():
    if(ct<=25):
        param.requires_grad = False
    else:
        param.requires_grad = True
    ct = ct + 1

#checkpoint = torch.load('backup_prev/checkpoint_suppress_best.tar')
#net.load_state_dict(checkpoint['state_dict'])
net.cuda()
learning_rate = 1e-6
numTopics = 21
backClass = 20
optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, net.parameters()), lr=learning_rate)


# In[ ]:


net.conv16 = nn.Conv2d(512, 512, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))
net.conv18 = nn.Conv2d(512, numTopics, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))


# In[ ]:


batchSize = 10
loadSize = 256
fineSize = 224
docLength = 56
Images_ROOT = 'images'


mean=[0.485, 0.456, 0.406]
std=[0.229, 0.224, 0.225]
import torchvision.datasets as datasets
data_transform = transforms.Compose([
            ScaleNonPreserving(224),
            transforms.ToTensor(),
        ])
data =  datasets.ImageFolder(Images_ROOT, data_transform)
train_loader = torch.utils.data.DataLoader(data, batch_size=batchSize, shuffle=True,
        num_workers=1, pin_memory=True)

numImages = len(data.imgs)
weights_fg = np.arange(docLength**2)
weights_fg = .996**weights_fg
weights_bg = np.arange(docLength**2)
weights_bg = .98**weights_bg
weights_fg_sum = weights_fg.sum()
weights_bg_sum = weights_bg.sum()


batch = torch.Tensor(batchSize,3, fineSize, fineSize).cuda()
batchScaled = np.zeros((batchSize, docLength, docLength,3), dtype=np.uint8)
batchL = []
dKL_dw = torch.zeros(batchSize, numTopics, docLength, docLength).cuda()
dce_dw = torch.zeros(batchSize, numTopics, docLength, docLength).cuda()
dce_dw2 = torch.zeros(batchSize, numTopics, docLength, docLength).cuda()


# In[ ]:


def getLocalDis(img, wordDis):
    # An enclosing function that takes pixel level distribution as input and computes local averages
    # for each pixel 
    # Converts the pixel level distribution to numpy for interfacing with the averaging code
    # Converts the local average distribution back to Tensor
    
    #Initialize the averaging code
    

    sigma_dis = 60*docLength/fineSize
    sigma_col = 13
    #Convert wordDis to numpy array
    tempQ = wordDis.cpu().numpy().reshape(numTopics,-1)

    d = dcrf.DenseCRF2D(img.shape[1], img.shape[0], numTopics)
    d.setUnaryEnergy(tempQ)
    d.addPairwiseGaussian(sxy=(3, 3), compat=3, kernel=dcrf.DIAG_KERNEL,
                  normalization=dcrf.NORMALIZE_SYMMETRIC)
    d.addPairwiseBilateral(sxy=(sigma_dis, sigma_dis), srgb=(sigma_col, sigma_col, sigma_col), rgbim=img,
                   compat=10,
                   kernel=dcrf.DIAG_KERNEL,
                   normalization=dcrf.NORMALIZE_SYMMETRIC)

    
    #Get Local average
    localDis, tmp1, tmp2 = d.startInference()
    localDis = eigen.matrixXf(np.array(tempQ).copy(order = 'C'))
    d.stepInference(localDis, tmp1, tmp2)

    #Convert localDis to Tensor
    localDis = np.array(localDis)
    localDis = torch.from_numpy(localDis.reshape(numTopics,docLength,docLength)).cuda()
    return localDis


# In[ ]:


def invert_permutation(p):
    '''The argument p is assumed to be some permutation of 0, 1, ..., len(p)-1. 
    Returns an array s, where s[i] gives the index of i in p.
    '''
    s = np.empty(p.size, p.dtype)
    s[p] = np.arange(p.size)
    return s


# In[ ]:


def fDx(batch, batchScaled, batchL):
    batchV = Variable(batch)
    output = net(batchV)

    KLD = 0
    class_error_fg = 0
    class_error_bg = 0
    dKL_dw.zero_()
    dce_dw.zero_()
    dce_dw2.zero_()
    for j in range(0, batchSize):
        global wordDis, localDis
        img = batchScaled[j]
        wordDis = output[j].data
        labelImage = batchL[j]
        foreGProb = wordDis[labelImage].cpu().numpy()
        backGProb = wordDis[backClass].cpu().numpy()
        #Ensuring classification performance
        #For each object in the image, find the top 10% pixels
        if(labelImage!=14):
            weights_fg = weights_fg_1
            weights_fg_sum = weights_fg_1_sum
        else:
            weights_fg = weights_fg_2
            weights_fg_sum = weights_fg_2_sum
        inds = np.argsort(foreGProb.reshape(docLength**2))[::-1]
        reverse_inds = invert_permutation(inds)
        weights_fg_reversed = weights_fg[reverse_inds].reshape(docLength,docLength)
        class_prob = (foreGProb*weights_fg_reversed).sum()/weights_fg_sum
        class_error_fg = class_error_fg + math.log(class_prob)*-1
        dce_dw[j].zero_()
        dce_dw[j][labelImage] = torch.from_numpy(weights_fg_reversed/class_prob*-1)

        if(labelImage!=backClass):
            inds = np.argsort(backGProb.reshape(docLength*docLength))[::-1]
            reverse_inds = invert_permutation(inds)
            weights_bg_reversed = weights_bg[reverse_inds].reshape(docLength,docLength)
            class_prob = (backGProb*weights_bg_reversed).sum()/weights_bg_sum
            class_error_bg = class_error_bg + math.log(class_prob)*-1
            dce_dw[j][backClass] = torch.from_numpy(weights_bg_reversed/class_prob*-1)
        
        #Ensuring smoothness
        localDis = getLocalDis(img, wordDis)
        #localLogDis -= localLogDis.max()
        #localDis = torch.exp(localLogDis)
        #localDis = localDis/localDis.sum(0).expand_as(localDis)
        wordByDis = torch.log(wordDis) - torch.log(localDis)
        KLD = KLD + torch.mul(wordDis, wordByDis).sum()
        dKL_dw[j] = (wordByDis + 1)

    weighting_factor =  (weights_fg_sum + weights_bg_sum)/docLength**2
        
    net.zero_grad()
    output.backward(0*dKL_dw+dce_dw)
    optimizer.step()
    return KLD, class_error_fg, class_error_bg


# In[ ]:


weights_fg_1 = np.arange(docLength**2)
weights_fg_1 = .998**weights_fg_1
weights_bg = np.arange(docLength**2)
weights_bg = 0.998**weights_bg
net.train()
weights_fg_2 = np.arange(docLength**2)
weights_fg_2 = .9985**weights_fg_2
net.cuda();
weights_total = np.arange(docLength**2)[::-1]
weights_total = .996**weights_total
weights_total[0:56*46] = 0
#weights_fg[1:math.ceil((docLength**2)/5)] = 1
#weights_fg[math.ceil((docLength**2)/5):] = 0
#weights_bg[1:math.ceil((docLength**2)/3)] = 1
#weights_bg[math.ceil((docLength**2)/3):] = 0
weights_fg_1_sum = weights_fg_1.sum()
weights_bg_sum = weights_bg.sum()
weights_fg_2_sum = weights_fg_2.sum()
#optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, net.parameters()), lr=learning_rate)

KLD =0
class_error_fg = 0
class_error_bg = 0
for epoch in range(0,1):
    for i, (img, label) in enumerate(train_loader):
        batchL = []
        for j in range(0, batchSize):
            batchScaled[j] = scaleImage( toPilImage(img[j]), docLength)
            batch[j] = normalize(img[j], mean, std)
            batchL.append(label[j])
            
        bKL, bcef, bceb = fDx(batch, batchScaled, batchL)
        KLD = KLD + bKL
        class_error_fg = class_error_fg + bcef
        class_error_bg = class_error_bg + bceb
        if(i%20==0):
            print(str(i)+' '+str(KLD)+' '+str(class_error_fg) + ' ' + str(class_error_bg))
            KLD =0
            class_error_fg = 0
            class_error_bg = 0
        if(i%200==0 and i>0):
            break


# In[ ]:
def test(loc):
    net.cuda();
    iterLoader = iter(train_loader)
    img, label = iterLoader.next()
    batchL = []
    batchScaled = np.zeros((
        batchSize, docLength, docLength,3), dtype=np.uint8)
    for j in range(0, batchSize):
        batchScaled[j] = scaleImage( toPilImage(img[j]), docLength)
        batch[j] = normalize(img[j], mean, std)
        batchL.append(label[j])
    batchV = Variable(batch)
    output = net(batchV)


    # In[ ]:


    label = batchL[loc]
    print(label)
    labelWeights = output[loc].data.cpu().numpy().sum(2).sum(1)
    for topic in range(0, numTopics):
        if labelWeights[topic]>0 and (topic==label or topic==label):
            print(topic)
            print(labelWeights[topic])
            displayMaskedImage(batchScaled[loc], output[loc][topic])
            plt.figure()
    plt.imshow(batchScaled[loc])



# In[ ]:


def displayMaskedImage(img, mask):
    if ~isinstance(img, np.ndarray):
        cur_data = np.asarray(img)
    else:
        cur_data = img

    if isinstance(mask, Variable):
        mask = mask.data.cpu().numpy()
    elif ~isinstance(mask, np.ndarray):
        mask = mask.cpu().numpy()
    
    #mask = (mask-mask.min())/(mask.max()-mask.min())
    mask = np.tile(mask.reshape(img.shape[0],img.shape[1],1), (1,1,3) )
    maskedImage = mask*cur_data
    plt.imshow(maskedImage/maskedImage.max())


# In[ ]:


checkpoint = {'state_dict':net.cpu().state_dict(),
              'optimizer':optimizer.state_dict()}


# In[ ]:


torch.save(checkpoint, 'checkpoint_cpu.tar')

